/*
*	nosey.c - rodent diagnostic agent :-)
*		 (c) 1993 Migrant Computing Services
*
*	gil@netcom.com 11Aug93 -
*	The mouse packet interpretation code is from the XFREE distribution
*	in the file x385Io.c (see the copyright by Thomas Roell below).
*	I've modified the mouse event proccessing by adding delayed queuing.
*	I have a Logitech MouseMan which needs it for proper middle
*	button handling. I also added the rest of the diagnostic code :-) 
*/

/*
 * Copyright 1990,91 by Thomas Roell, Dinkelscherben, Germany.
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of Thomas Roell not be used in
 * advertising or publicity pertaining to distribution of the software without
 * specific, written prior permission.  Thomas Roell makes no representations
 * about the suitability of this software for any purpose.  It is provided
 * "as is" without express or implied warranty.
 *
 * THOMAS ROELL DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 * EVENT SHALL THOMAS ROELL BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 *
 * $Header: /sda1/ports/nosey/RCS/nosey.c,v 1.2 1993/08/11 11:42:27 root Exp root $
 */

#include <stdio.h>
#include <termios.h>
#include <sys/types.h>
#include <sys/time.h>	/* for FD_*() macros */
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <signal.h>


/* mouse types */
#define P_MS    0                     /* Microsoft */
#define P_MSC   1                     /* Mouse Systems Corp */
#define P_MM    2                     /* MMseries */
#define P_LOGI  3                     /* Logitech */
#define P_BM    4                     /* BusMouse ??? */
#define P_END   4                     /* make equal to large of mouse types # */


/* mouse action types */
typedef enum tagMSEACT {
	MOUSE_INIT, MOUSE_ON, MOUSE_OFF, MOUSE_CLOSE
} MSEACT;

typedef enum tagKBDACT {
	KBD_INIT, KBD_ON, KBD_OFF, KBD_CLOSE
} KBDACT;

/* mouse attributes */
static int mseFd = -1;
static char *mseDevice = "/dev/ttys1";
static int mseType = P_MS;
static int baudRate = B1200;
static int sampleRate = 0;
static int lastButtons = 0;
/* int emulateState = 0; */

/* variables for motion and position tracking */
static int xhot = 1, yhot = 1;
static int xhot_min = 0, xhot_max = 79;
static int yhot_min = 0, yhot_max = 24;
static int x_decay = 6, y_decay = 6;
static int x_thresh = 6, y_thresh = 6;
static int dx_accum = 0, dy_accum = 0;

/* maximum mouse polling counter/timer for mouse event queueing */
#define MAX_MOUSE_POLL 250

/* variables for mouse event queuing */
static int qmse_active = 0;
static int qmse_buttons = 0;
static int qmse_dx = 0;
static int qmse_dy = 0;
static int qmse_aging = 0;
static int qmse_timer = 0;


static void
PostMseEvent(int buttons, int dx, int dy)
{
	/* gen new position */
	if ((dx + dx_accum > x_thresh) || (dx + dx_accum < -x_thresh))
	{
		xhot += (dx + dx_accum)/x_decay;
		dx_accum = 0;
	}
	else
		dx_accum += dx;
	if ((dy + dy_accum > y_thresh) || (dy + dy_accum < -y_thresh))
	{
		yhot += (dy + dy_accum)/y_decay;
		dy_accum = 0;
	}
	else
		dy_accum += dy;

	if ((xhot < xhot_min) || (xhot > xhot_max))
		if (xhot < xhot_min)
			xhot = xhot_min;
		else
			xhot = xhot_max;
	if ((yhot < yhot_min) || (yhot > yhot_max))
		if (yhot < yhot_min)
			yhot = yhot_min;
		else
			yhot = yhot_max;
#ifdef SIMPLEMOUSE
	printf("Btn=%d dx=%d dy=%d\n", buttons, dx, dy);
#else
	printf("\rButtonsDown=[");
	if (buttons & 0x4) printf("left,"); else printf("    ,");
	if (buttons & 0x2) printf("middle,"); else printf("      ,");
	if (buttons & 0x1) printf("right]"); else printf("     ]");
	printf(" xhot=%02.2d, yhot=%02.2d", xhot, yhot);
	printf(" dx=%+03.3d, dy=%+03.3d\r", dx, dy);
	fflush(stdout);
#endif /* SIMPLEMOUSE */
}


static void
SetSpeed(int old, int new, unsigned short cflag)
{
	struct termios t;
	char           *c;

	if (tcgetattr(mseFd, &t) != 0)
	{
		fprintf(stderr, "Can't get attribute for mousedev!!\n");
		exit(0);
	}

	t.c_iflag &= ~(BRKINT | IGNPAR | PARMRK | INPCK | ISTRIP
		| INLCR | IGNCR | ICRNL | IXON);
	t.c_iflag |= IGNBRK | IGNPAR;
	t.c_oflag &= ~(OPOST);
	t.c_lflag &= ~(ECHO | ECHOE | ECHOK | ECHONL |
		ICANON | ISIG | NOFLSH | TOSTOP);
	t.c_cflag &= ~(CBAUD | CSIZE | CSTOPB | CREAD | PARENB \
		| PARODD | HUPCL | CLOCAL);
	t.c_cflag |= cflag;
	t.c_cc[VMIN] = 0;
	t.c_cc[VTIME] = 5;

	if (cfsetispeed(&t, old) == -1 || cfsetospeed(&t, old) == -1)
	{
		fprintf(stderr, "Can't set mouse old I/O speed\n");
		exit(0);
	}

	if (tcsetattr(mseFd, TCSANOW, &t) == -1)
	{
		fprintf(stderr, "Can't set mouse attributes\n");
		exit(0);
	}

	if (cfsetispeed(&t, new) == -1 || cfsetospeed(&t, new) == -1)
	{
		fprintf(stderr, "Can't set mouse new I/O speed\n");
		exit(0);
	}

	switch(new)
	{
		case B9600:  c = "*q";  break;
		case B4800:  c = "*p";  break;
		case B2400:  c = "*o";  break;
		case B1200:
		default:    c = "*n";  break;
	}

	write(mseFd, c, 2);  
	usleep(100000);

	if (tcsetattr(mseFd, TCSANOW, &t) == -1)
	{
		fprintf(stderr, "Can't set new mouse attributes\n");
		exit(0);
	}

	tcflush(mseFd, TCIFLUSH);	/* flush any pending input */

#ifdef TCMOUSE
	(void) ioctl(mseFd, TCMOUSE, 1);  
#endif
}



/*
* MseProc --
*      Handle the initialization, etc. of a mouse
*/
static int
MseProc(MSEACT what)
{
	unsigned char                map[4];

	static unsigned short cflag[5] =
	{
		(CS7                   | CREAD | CLOCAL | HUPCL ),   /* MicroSoft */
		(CS8 | CSTOPB          | CREAD | CLOCAL | HUPCL ),   /* MouseSystems */
		(CS8 | PARENB | PARODD | CREAD | CLOCAL | HUPCL ),   /* MMSeries */
		(CS8 | CSTOPB          | CREAD | CLOCAL | HUPCL ),   /* Logitech */
		0,                                                   /* BusMouse */
	};

	switch (what)
	{
		case MOUSE_INIT: 
			break;

		case MOUSE_ON:
			if ((mseFd = open(mseDevice, O_RDWR | O_NDELAY)) < 0)
			{
				fprintf(stderr,"Cannot open mouse");
				return (-1);
			}

			if (mseType != P_BM) 
			{
				SetSpeed(B9600, baudRate, cflag[mseType]);
				SetSpeed(B4800, baudRate, cflag[mseType]);
				SetSpeed(B2400, baudRate, cflag[mseType]);
				SetSpeed(B1200, baudRate, cflag[mseType]);

				if (mseType == P_LOGI)
				{
					write(mseFd, "S", 1);
					SetSpeed(baudRate, baudRate, cflag[P_MM]);
				}

				if      (sampleRate <=   0)  write(mseFd, "O", 1);
				else if (sampleRate <=  15)  write(mseFd, "J", 1);
				else if (sampleRate <=  27)  write(mseFd, "K", 1);
				else if (sampleRate <=  42)  write(mseFd, "L", 1);
				else if (sampleRate <=  60)  write(mseFd, "R", 1);
				else if (sampleRate <=  85)  write(mseFd, "M", 1);
				else if (sampleRate <= 125)  write(mseFd, "Q", 1);
				else                                  write(mseFd, "N", 1);
			}


			lastButtons = 0;
			/* emulateState = 0; */
			break;

		case MOUSE_OFF:
		case MOUSE_CLOSE:
			if (mseType == P_LOGI)
			{
				write(mseFd, "U", 1);
				SetSpeed(baudRate, B1200, cflag[P_LOGI]);
			}
			close(mseFd);
			usleep(300000);
			break;

	}

	return 0;
}


static void
MseClose(void)
{
	MseProc(MOUSE_CLOSE);
}


static void
QInMseEvent(int buttons, int dx, int dy)
{
	qmse_buttons = buttons;
	qmse_dx = dx;
	qmse_dy = dy;
	qmse_active = 1;
}

static void
QOutMseEvent()
{
	if (qmse_active)
	{
		PostMseEvent(qmse_buttons, qmse_dx, qmse_dy);
		qmse_active = 0;
	}
}

int
QPushMseEvent(int buttons, int dx, int dy)
{
	int qwasempty;

	qwasempty = !qmse_active;
	QOutMseEvent();
	QInMseEvent(buttons, dx, dy);
	return(qwasempty);
}

void
QModMseEvent(int buttons, int dx, int dy)
{
	if (qmse_active)
	{
		qmse_buttons = buttons; /* update buttons */
		QOutMseEvent();
		return;
	}
	/* else yuk, got a modify prev event with no prev event */
	/* PostMseEvent(buttons, dx, dy); */ /* fast mode */
	QInMseEvent(buttons, dx, dy);
	QOutMseEvent();
	return;
}

int
KbdEvents()
{
	int i;
	if (toupper(getchar()) == 'Q')
		raise(SIGINT);
	return(0);
}

/*
* MseEvents --
*      Read the new events from the device, and pass them to the eventhandler.
*/
int
MseEvents()
{
	unsigned char        rBuf[64];
	int                  i,nBytes, buttons, dx, dy;
	static int           pBufP = 0;
	static unsigned char pBuf[8];
	int pollagain = 0;

	static unsigned char proto[5][5] = {
		/*  hd_mask hd_id   dp_mask dp_id   nobytes */
		{ 	0x40,	0x40,	0x40,	0x00,	3 	},  /* MicroSoft */
		{	0xf8,	0x80,	0x00,	0x00,	5	},  /* MouseSystems */
		{	0xe0,	0x80,	0x80,	0x00,	3	},  /* MMSeries */
		{	0xe0,	0x80,	0x80,	0x00,	3	},  /* Logitech */
		{	0xf8,	0x80,	0x00,	0x00,	5	},  /* BusMouse */
	};

	nBytes = read(mseFd, (char *)rBuf, sizeof(rBuf));
	if ( nBytes <= 0 )
	{
		if ( !qmse_aging )
		{
			qmse_aging = 1;
			qmse_timer = 0;
			return(1);
		}
		else
		{
			if ( qmse_timer < MAX_MOUSE_POLL )
			{
				qmse_timer++;
				return(1);
			}
			qmse_aging = 0;
			qmse_timer = 0;
		}
		QOutMseEvent();
		return(0);
	}
	else
	{
		qmse_aging = 0;
		qmse_timer = 0;
	}
	
	

	for ( i=0; i < nBytes; i++)
	{
		/*
		* Hack for resyncing: We check here for a package that is:
		*  a) illegal (detected by wrong data-package header)
		*  b) invalid (0x80 == -128 and that might be wrong for MouseSystems)
		*  c) bad header-package
		*
		* NOTE: b) is a voilation of the MouseSystems-Protocol, since values of
		*       -128 are allowed, but since they are very seldom we can easily
		*       use them as package-header with no button pressed.
		*/
		if (pBufP != 0 && 
			((rBuf[i] & proto[mseType][2]) != proto[mseType][3]
			|| rBuf[i] == 0x80))
		{
			pBufP = 0;          /* skip package */
		}

		if (pBufP == 0 &&
			(rBuf[i] & proto[mseType][0]) != proto[mseType][1])
		{
			/*
			* Hack for Logitech MouseMan Mouse - Middle button
			*
			* Unfortunately this mouse has variable length packets: the standard
			* Microsoft 3 byte packet plus an optional 4th byte whenever the
			* middle button status changes.
			*
			* We have already processed the standard packet with the movement
			* and button info.  Now post an event message with the old status
			* of the left and right buttons and the updated middle button.
			*/
			if (mseType == P_MS && (rBuf[i] == 0x20 || rBuf[i] == 0))
			{
				buttons = ((int)(rBuf[i] & 0x20) >> 4)
					| (lastButtons & 0x05);
				/* PostMseEvent(buttons, 0, 0); */
				QModMseEvent(buttons, 0, 0);
				/* pollagain = 0; */
				lastButtons = buttons;
			}

			continue;            /* skip package */
		}

		pBuf[pBufP++] = rBuf[i];
		if (pBufP != proto[mseType][4]) continue;

		/*
		* assembly full package
		*/
		switch(mseType)
		{
			case P_MS:              /* Microsoft */
				buttons = (lastButtons & 2)
					| ((int)(pBuf[0] & 0x20) >> 3)
					| ((int)(pBuf[0] & 0x10) >> 4);
				dx = (char)(((pBuf[0] & 0x03) << 6) | (pBuf[1] & 0x3F));
				dy = (char)(((pBuf[0] & 0x0C) << 4) | (pBuf[2] & 0x3F));
				break;

			case P_MSC:             /* Mouse Systems Corp */
				buttons = (~pBuf[0]) & 0x07;
				dx =    (char)(pBuf[1]) + (char)(pBuf[3]);
				dy = - ((char)(pBuf[2]) + (char)(pBuf[4]));
				break;

			case P_MM:              /* MM Series */
			case P_LOGI:            /* Logitech Mice */
				buttons = pBuf[0] & 0x07;
				dx = (pBuf[0] & 0x10) ?   pBuf[1] : - pBuf[1];
				dy = (pBuf[0] & 0x08) ? - pBuf[2] :   pBuf[2];
				break;

			case P_BM:              /* BusMouse */
				buttons = (~pBuf[0]) & 0x07;
				dx =   (char)pBuf[1];
				dy = - (char)pBuf[2];
				break;
		}

		/* PostMseEvent(buttons, dx, dy); */
		if (QPushMseEvent(buttons, dx, dy))
			pollagain = 1;
		lastButtons = buttons;
		pBufP = 0;
	}
	/* return(pollagain); */
	return(1);
}

void
CloseAll(int why)
{
	puts("\nThanx NOSEY.");
	exit();
}

/*
* KLUDGE: Checks for anything set in the given fd_set structure.
*	We know this works on linux because there is a contiguous set of
*	longs that hold the bit values. This ought to be a macro along with
*	the other FD_* macros. It may have to be enclosed in a 
*	#if !defined(_POSIX_SOURCE).....#endif
*/
#ifndef FD_ANYSET
int
FD_ANYSET(fd_set *fdstuff)
{
	register lencheck;
	unsigned char *tptr;

	tptr = (unsigned char *)fdstuff;
	lencheck = sizeof(fd_set);
	for(; lencheck >= 0; lencheck--)
		if ( *tptr )
			return(1);
	return(0);
}
#endif


int
KbdProc(KBDACT what)
{
	struct termios kbdattr;

	tcgetattr(0, &kbdattr);
	switch(what)
	{
		case KBD_INIT:
			kbdattr.c_lflag &= ~(ICANON | ECHO);
			kbdattr.c_cc[VMIN] = 0;
			kbdattr.c_cc[VTIME] = 0;
			break;
		case KBD_ON:
			kbdattr.c_lflag |= ECHO;
			break;
		case KBD_OFF:
			kbdattr.c_lflag &= ~ECHO;
			break;
		case KBD_CLOSE:
			kbdattr.c_lflag |= ICANON | ECHO;
			kbdattr.c_cc[VMIN] = 1;
			kbdattr.c_cc[VTIME] = 0;
			break;
		/* enum use restrict default cases */
	}
	tcsetattr(0, TCSANOW, &kbdattr);
}

KbdClose(int why)
{
	KbdProc(KBD_CLOSE);
}

char *explaindisplay[] = {
	"", 
	"This is NOSEY, a rodent diagnostic program.",
	"The one line dynamic display below is used to track nosey.",
	"\tBracketed items track [left,middle,right] button clicks.",
	"\txhot tracks left-right 'curses' position from 0 to 79.",
	"\tyhot tracks up-down 'curses' positions from 0 to 24.",
	"\tdx tracks left-right signed delta motions.",
	"\tdy tracks up-down signed delta motions.",
	"",
	"Nosey was Angela Marie R.'s rat.",
	"This checks other rodents too :-)",
	"",
	"Set rodent loose! Press q or Q to quit.",
	NULL
};

char *explainusage[] = {
	"", "",
	"Usage: nosey <maker_code> <device_name> <baudrate> <samplerate>",
	"This is NOSEY, a rodent diagnostic program.",
	"It needs a maker or emulation:",
	"\tMicrosoft - use 0",
	"\tMouseSystems - use 1",
	"\tMMSeries - use 2",
	"\tLogitech - use 3",
	"\tLogitech MouseMan - use 0  (again)",
	"\tBusmouse - use 4",
	"It needs a device: like /dev/mouse or /dev/ttys1 or /dev/ttys2...",
	"It needs a baud rate: 1200, 2400, 4800, 9600.",
	"It needs a sample rate: 0 to N (like 150).",
	"",
	"Nosey was Angela Marie R.'s rat.",
	"This checks other rodents too :-)",
	"",
	NULL
};


void
Explain(char *explainthis[])
{
	register index;
	for(index=0; explainthis[index]; index++)
		puts(explainthis[index]);
	fflush(stdout);
	return;
}

main(int argc, char *argv[])
{
	fd_set fdvar, polldevs;
	int numready;

	if (argc < 5)
	{
		fprintf(stderr, "Not enough arguments.\n");
		Explain(explainusage);
		exit(-1);
	}
	mseType = atoi(argv[1]);
	if (mseType < 0 || mseType > P_END)
	{
		fprintf(stderr, "Invalid manufacturuer code: %s\n", argv[1]);
		Explain(explainusage);
		exit(-1);
	}
	mseDevice = argv[2];
	switch( atoi(argv[3]) )
	{
		case 1200:  baudRate = B1200; break;
		case 2400:  baudRate = B2400; break;
		case 4800:  baudRate = B4800; break;
		case 9600:  baudRate = B9600; break;
		default:
			fprintf(stderr, "Invalid baud rate: %s\n", argv[3]);
			Explain(explainusage);
			exit(-1);
	}
	sampleRate = atoi(argv[4]);
	fprintf(stderr, "<Internal info: %d %s %d %d>\n",
		mseType, mseDevice, baudRate, sampleRate);

	/* prepare mouse device */
	MseProc(MOUSE_INIT);
	MseProc(MOUSE_ON);
	atexit(MseClose);

	/* prepare kbd device */
	KbdProc(KBD_INIT);
	KbdProc(KBD_OFF);	/* don't echo keys */
	atexit(KbdClose);

	signal(SIGINT, CloseAll);

	Explain(explaindisplay);

	/* prepare event gathering */
	FD_ZERO(&polldevs);
	for(;;)
	{ /* look for keyboard and mouse events */
		if ( FD_ANYSET(&polldevs) )
		{ /* check for devices to be polled first */
			if ( FD_ISSET(mseFd, &polldevs) )
			{
				if ( MseEvents() )
					FD_SET(mseFd, &polldevs);
				else
					FD_CLR(mseFd, &polldevs);
				FD_SET(0, &polldevs); /* check for kbd too */
			}
			if ( FD_ISSET(0, &polldevs) )
			{
				if ( KbdEvents() )
					FD_SET(0, &polldevs);
				else
					FD_CLR(0, &polldevs);
			}
			if ( FD_ANYSET(&polldevs) )
				continue; /* keep polling till no polling */
		}

		/* prepare mouse select descriptors */
		FD_ZERO(&fdvar);
		FD_SET(mseFd, &fdvar);

		/* prepare kbd select descriptors */
		FD_SET(0, &fdvar);

		/*
		*  note: 1st arg of select is "greatest numbered file
		*	 descriptor + 1" to monitor, or/upto max of 256.
		*/
		numready = select(mseFd + 1, &fdvar,
				 (fd_set *)NULL, (fd_set *)NULL, NULL);
		if ( numready < 0 )
		{
			fprintf(stderr, "Oops, select booboo!\n");
			continue;
		}
		else if (FD_ISSET(mseFd, &fdvar))
		{
			if ( MseEvents() )
				FD_SET(mseFd, &polldevs);
			else
				FD_CLR(mseFd, &polldevs);
		}
		else if (FD_ISSET(0, &fdvar))
		{
			if ( KbdEvents() )
				FD_SET(0, &polldevs);
			else
				FD_CLR(0, &polldevs);
		}
		/* else continue; */
	}
}
